#include "bully.h"
#include "bullyparams.h"
#include "bullybuffer.h"
#include "bullybuffermanager.h"
#include "bullymath.h"

#include <cv.h>
#include <profiling.h>
#include <time.h>
#include <iostream>
#include <sstream>

// Minimal/dummy modules
#include "bullymoduleprep_default.h"
#include "bullymoduletrack_default.h"
#include "bullymoduleanal_default.h"

// TDT4725 Depth Study
#include "bullymoduleprep_tdt4725.h"
#include "bullymoduletrack_tdt4725.h"
#include "bullymoduleanal_tdt4725.h"

// New modules
#include "bullymoduleprep_hsv.h"
#include "bullymoduleprep_cncc_hsl.h"
#include "bullymoduleprep_cncc_rgb.h"
#include "bullymoduleprep_cncc_hsv.h"
#include "bullymoduletrack_reading.h"
#include "bullymoduletrack_master.h"
#include "bullymoduleanal_notstanding_shadow.h"
#include "bullymoduleanal_notstanding.h"
#include "bullymoduleanal_chase.h"
#include "bullymoduleanal_debug.h"

using namespace std;

void Bully::System::createModules(){
	modPreprocess["default"] = new ModulePreprocess_Default( *buffer, params );
	modTrack["default"] = new ModuleTrack_Default( *buffer, params );
	modAnalyze["default"] = new ModuleAnalyze_Default( *buffer, params );

	modPreprocess["tdt4725"] = new ModulePreprocess_Tdt4725( *buffer, params );
	modTrack["tdt4725"] = new ModuleTrack_Tdt4725( *buffer, params );
	modAnalyze["tdt4725"] = new ModuleAnalyze_Tdt4725( *buffer, params );

	modPreprocess["hsv"] = new ModulePreprocess_HSV( *buffer, params );
	modPreprocess["cncc"] = new ModulePreprocess_CNCC_hsL( *buffer, params );
//	modPreprocess["cncc_rgb"] = new ModulePreprocess_CNCC_RGB( *buffer, params );
//	modPreprocess["cncc_hsv"] = new ModulePreprocess_CNCC_HSV( *buffer, params );

	modTrack["reading"] = new ModuleTrack_Reading( *buffer, params );
	modTrack["master"] = new ModuleTrack_Master( *buffer, params );

	modAnalyze["notstandings"] = new ModuleAnalyze_NotStanding_Shadow( *buffer, params );
	modAnalyze["notstandingh"] = new ModuleAnalyze_NotStanding( *buffer, params );
	modAnalyze["chase"] = new ModuleAnalyze_Chase( *buffer, params );
	modAnalyze["debug"] = new ModuleAnalyze_Debug( *buffer, params );
}

Bully::System::System(){
	init();
}
Bully::System::~System(){
	destroy();
}
void Bully::System::init(){
	initMath();
	buffer = new BufferManager();
	createModules();
}
void Bully::System::destroy(){
	destroyModules();
	delete buffer; buffer = 0;
}
void Bully::System::printModules(){
	std::map<std::string, ModulePreprocess*>::iterator prepit;
	std::map<std::string, ModuleTrack*>::iterator trackit;
	std::map<std::string, ModuleAnalyze*>::iterator analit;

	cout << "Available preprocessing modules:" << endl;
	for( prepit = modPreprocess.begin(); prepit != modPreprocess.end(); prepit++ ){
		if( prepit->first == params.getString("modulePreprocess") )
			cout << " * " << prepit->first << endl;
		else
			cout << "   " << prepit->first << endl;
	}
	cout << "Available tracking modules:" << endl;
	for( trackit = modTrack.begin(); trackit != modTrack.end(); trackit++ ){
		if( trackit->first == params.getString("moduleTrack") )
			cout << " * " << trackit->first << endl;
		else
			cout << "   " << trackit->first << endl;
	}
	cout << "Available analysis modules:" << endl;
	for( analit = modAnalyze.begin(); analit != modAnalyze.end(); analit++ ){
		bool inUse = false;
		for( int i = 0; i < (int)params.valueCount("moduleAnalyze"); i++ ){
			if( analit->first == params.getString("moduleAnalyze", i) ){
				inUse = true;
				break;
			}
		}
		if( inUse )
			cout << " * " << analit->first << endl;
		else
			cout << "   " << analit->first << endl;
	}
}
void Bully::System::destroyModules(){
	std::map<std::string, ModulePreprocess*>::iterator prepit;
	std::map<std::string, ModuleTrack*>::iterator trackit;
	std::map<std::string, ModuleAnalyze*>::iterator analit;

	for( prepit = modPreprocess.begin(); prepit != modPreprocess.end(); prepit++ )
		delete prepit->second;
	for( trackit = modTrack.begin(); trackit != modTrack.end(); trackit++ )
		delete trackit->second;
	for( analit = modAnalyze.begin(); analit != modAnalyze.end(); analit++ )
		delete analit->second;
}

//************************************************************************
// NB: all frames are 24 bit BGR!
//************************************************************************
void Bully::System::init( int w, int h, uchar* refFrame ){
	buffer->init(w, h);
	buffer->createBuffer( w, h, 3, DEPTH_8U, "background_rgb" );
	buffer->copy(refFrame, "background_rgb");
}
Bully::System::Assessment Bully::System::assessSituation( uchar* frame ){
	PROFILE_BEGIN_FUNC;
	checkParams();

	assert(buffer->exists("background_rgb"));
	int w = buffer->get("background_rgb")->w();
	int h = buffer->get("background_rgb")->h();
	// createBuffer returns at once if the buffer already exists.
	buffer->createBuffer( w, h, 3, DEPTH_8U, "input_rgb" );
	buffer->copy(frame, "input_rgb");

	// START ASSESSMENT

	// Reset output
	buffer->get("output")->setZero();

	ModulePreprocess *preprocess =
		modPreprocess[params.getString("modulePreprocess")];
	ModuleTrack *track =
		modTrack[params.getString("moduleTrack")];

	if( (int)params.get("printFrameCount") != 0 )
		cout << endl << preprocess->frameCount() << ": ";

	preprocess->process();
	track->process( preprocess->getBlobList() );

	// BEGIN MULTITHREADED

	Assessment ass;
	for( int i = 0; i < (int)params.valueCount("moduleAnalyze"); i++ ){
		ModuleAnalyze *analyze = modAnalyze[params.getString("moduleAnalyze", i)];
		ModuleAnalyze::Result res = analyze->process(track->getObjectList());

		for( int a = 0; a < res.alarms.size(); a++ ){
			ass.alarms.push_back( Alarm(res.alarms[a].area, res.alarms[a].desc) );
		}
	}

	// END MULTITHREADED -- SYNCH POINT, BARRIER
	//WaitForMultipleObjects(MAX_THREADS, hThread, TRUE, INFINITE);

	PROFILE_END_FUNC;
	return ass;
}
unsigned char * Bully::System::output(){
	return buffer->get("output")->data();
}
void Bully::System::handleInput( bool down, char ascii ){
	// Send input on to subsystems.
	modPreprocess[params.getString("modulePreprocess")]->handleInput(down, ascii);
	modTrack[params.getString("moduleTrack")]->handleInput(down, ascii);

	for( int i = 0; i < (int)params.valueCount("moduleAnalyze"); i++ )
		modAnalyze[params.getString("moduleAnalyze", i)]->handleInput(down, ascii);
}
void Bully::System::handleInput( bool down, int button, int x, int y ){
	// Send input on to subsystems.
	modPreprocess[params.getString("modulePreprocess")]->handleInput(down, button, x, y);
	modTrack[params.getString("moduleTrack")]->handleInput(down, button, x, y);

	for( int i = 0; i < (int)params.valueCount("moduleAnalyze"); i++ )
		modAnalyze[params.getString("moduleAnalyze", i)]->handleInput(down, button, x, y);
}
void Bully::System::checkParams(){
	bool modPrepExists = 1 == modPreprocess.count(params.getString("modulePreprocess"));
	if( !modPrepExists ){
		cout	<< "NB: Preprocessing module " << params.getString("modulePreprocess")
				<< " does not exist. Falling back to default." << endl;
		params.set("modulePreprocess", "default");
	}
	bool modTrackExists = 1 == modTrack.count(params.getString("moduleTrack"));
	if( !modTrackExists ){
		cout	<< "NB: Tracking module " << params.getString("moduleTrack")
				<< " does not exist. Falling back to default." << endl;
		params.set("moduleTrack", "default");
	}
	bool modAnalExists = 1 == modAnalyze.count(params.getString("moduleAnalyze"));
	if( !modAnalExists ){
		cout	<< "NB: Analysis module " << params.getString("moduleAnalyze")
				<< " does not exist. Falling back to default." << endl;
		params.set("moduleAnalyze", "default");
	}

	// Throw exception if any module was not found!
	if( !modPrepExists || !modTrackExists || !modAnalExists )
		throw string("A module was not found, exiting.");
}
Bully::Buffer* Bully::System::getBuffer(const char* name){
	return buffer->get(name);
}
