#include "bullybuffer.h"
#include "bullybuffermanager.h"
#include "bullymath.h"
#include "bullybgfile.h"
#include "eriksul.h"
#include <profiling.h>
#include <cv.h>
#include <assert.h>
#include <sstream>
using namespace std;


Bully::Buffer::Buffer( const char * name, int w, int h, int c, PixelDepth d){
	assert( c == 1 || c == 3 );
	init(name, w, h, c, d);
}
Bully::Buffer::~Buffer(){
	cvReleaseImage( &_iplImage );
}
void Bully::Buffer::init(){
	_width =_height = 0;
	_depth = DEPTH_8U;
	_channels = 0;
	_pitch = 0;
	_size = 0;
	_data = 0;
	_fdata = 0;
	_iplImage = 0;
}
void Bully::Buffer::init( const char * name, int w, int h, int c, PixelDepth d ){
	init();

	this->_name = name;
	_width = w;
	_height = h;
	_channels = c;
	_depth = d;
	int pixelFormat = 0;
	switch(_depth){
		case DEPTH_8U:	pixelFormat = IPL_DEPTH_8U; break;
		case DEPTH_32F: pixelFormat = IPL_DEPTH_32F; break;
		default: assert(0); break;
	}

	_iplImage = cvCreateImage( cvSize(w,h), pixelFormat, _channels );
	_data = (uchar*)_iplImage->imageData;
	_fdata = (float*)_iplImage->imageData;
	_pitch = _iplImage->widthStep;
	_size = _iplImage->imageSize;
	setZero();
}
void Bully::Buffer::copy( unsigned char * srcData ){
	assert(srcData);
	assert(pitch() == w()*3);
	assert(size() == w()*h()*3);
	// Assume srcData is in BGR format.
	memcpy(_data, srcData, w()*h()*3);
}
void Bully::Buffer::copy( Buffer * srcBuffer, Buffer * mask ){
	assert(srcBuffer);
	assert(w() == srcBuffer->w());
	assert(h() == srcBuffer->h());
	assert(channels() == srcBuffer->channels());
	assert(depth() == srcBuffer->depth());
	if( mask ){
		assert(mask->depth() == DEPTH_8U);
		cvCopy( srcBuffer->iplImage(), iplImage(), mask->iplImage() );
	}else{
		cvCopy( srcBuffer->iplImage(), iplImage() );
	}
}

// Color conversion  ---------------------------------------------------------
Bully::Buffer* Bully::Buffer::gray2rgb(Buffer *dest){
	assert(w() == dest->w());
	assert(h() == dest->h());
	assert(channels() == 1);
	assert(dest->channels() == 3);
	PROFILE_BEGIN_FUNC;
	cvCvtColor( iplImage(), dest->iplImage(), CV_GRAY2BGR );
	PROFILE_END_FUNC;
	return dest;
}
Bully::Buffer* Bully::Buffer::rgb2gray(Buffer *dest){
	assert(w() == dest->w());
	assert(h() == dest->h());
	assert(channels() == 3);
	assert(dest->channels() == 1);
	PROFILE_BEGIN_FUNC;
	cvCvtColor( iplImage(), dest->iplImage(), CV_BGR2GRAY );
	PROFILE_END_FUNC;
	return dest;
}
Bully::Buffer* Bully::Buffer::rgb2gray_erik(Buffer *dest){
	assert(w() == dest->w());
	assert(h() == dest->h());
	assert(channels() == 3);
	assert(dest->channels() == 1);
	assert(depth() == DEPTH_8U);
	assert(dest->depth() == DEPTH_8U);
	PROFILE_BEGIN_FUNC;

	unsigned char * rgb = data();
	unsigned char * gray = dest->data();

	int rgboff = 0;
	int grayoff = 0;

	for( int y = 0; y < h(); y++ ){
		for( int x = 0; x < w(); x++ ){
			unsigned char b = rgb[rgboff + x*3 + 0];
			unsigned char g = rgb[rgboff + x*3 + 1];
			unsigned char r = rgb[rgboff + x*3 + 2];
			// Fixed point luminance approx: (See http://en.wikipedia.org/wiki/YUV)
			// lum = min(abs(r * 2104 + g * 4130 + b * 802 + 4096 + 131072) >> 13,235);
			gray[grayoff + x] = min(abs(r * 2104 + g * 4130 + b * 802 + 4096 + 131072) >> 13,235);
		}

		rgboff += pitch();
		grayoff += dest->pitch();
	}
	PROFILE_END_FUNC;
	return dest;
}
Bully::Buffer* Bully::Buffer::rgb2hsv(Buffer *dest){
	assert( w() == dest->w() );
	assert( h() == dest->h() );
	assert( channels() == 3 && dest->channels() == 3 );
	assert( depth() == dest->depth() );
	PROFILE_BEGIN_FUNC;
	cvCvtColor( iplImage(), dest->iplImage(), CV_BGR2HSV );
	PROFILE_END_FUNC;
	return dest;
}
Bully::Buffer* Bully::Buffer::rgb2hsL(Buffer *dest){
	assert( w() == dest->w() );
	assert( h() == dest->h() );
	assert( channels() == 3 && dest->channels() == 3 );
	assert( depth() == DEPTH_8U );
	assert( dest->depth() == DEPTH_32F );
	PROFILE_BEGIN_FUNC;
	int width = w();
	int height = h();
	float div255 = 1.0f / 255.0f;
	float div2div255 = 1.0f / 510.0f;

	assert( pitch() == width * 3 );
	assert( dest->pitch() == width * 3 * 4 );

	uchar *rgbdata = data();
	float *hsLdata = dest->fdata();

	int H = 0;
	float S = 0.0f;
	for( int i = 0; i < width*height*3; i += 3 ){
		uchar b = rgbdata[i + 0];  // channelSeq = BGR
		uchar g = rgbdata[i + 1];
		uchar r = rgbdata[i + 2];

		uchar maxrgb = max(r,max(g,b));
		uchar minrgb = min(r,min(g,b));
		uchar diff = max(1,maxrgb-minrgb); // To avoid division by zero

		if( maxrgb == r && g >= b )
			H = 60 * (g-b) / diff + 0;
		else if( maxrgb == r && g < b )
			H = 60 * ((g-b) / diff) + 360;
		else if( maxrgb == g )
			H = 60 * ((b-r) / diff) + 120;
		else // if( maxrgb == b )
			H = 60 * ((r-g) / diff) + 240;

		// S = (maxrgb-minrgb)/255 = (maxrgb-minrgb)*(1/255);
		S = float(maxrgb-minrgb) * div255;

		// h and s will be euclidean coordinates, so they can be negative as well as positive.
		// The length of (h,s) should be S.
		hsLdata[i + 0] = Bully::cosf[H] * S;  // h
		hsLdata[i + 1] = Bully::sinf[H] * S;  // s
		// L = ((maxrgb+minrgb)/2)/255 = (maxrgb+minrgb)*(1/510);
		hsLdata[i + 2] = float(maxrgb + minrgb) * div2div255;  // L
	}
/*
		float div255 = 1.0f / 255.0f;
		float div2div255 = 1.0f / 510.0f;
		for( int y = 0; y < height; y++ ){
			for( int x = 0; x < width; x++ ){

				//*************************************************
				// USE SSE2 INTRINSICS TO SPEED UP
				// _mm_add_pd -> addition of two 64 doubles
				// _mm_div_pd
				// _mm_mul_pd
				// _mm_max_pd
				// _mm_sqrt_pd
				// emmintrin.h -> sse intrinsics
				// mmintrin.h -> mmx intrinsics
				// NB: possibly need to use the EMMS instruction:
				// http://msdn2.microsoft.com/en-us/library/ays9ef83.aspx
				//*************************************************

				uchar r = pix8u(x,y,0);
				uchar g = pix8u(x,y,1);
				uchar b = pix8u(x,y,2);
				uchar maxrgb = max(r,max(g,b));
				uchar minrgb = min(r,min(g,b));
				uchar diff = max(1,maxrgb-minrgb); // To avoid division by zero

				int H = 0;
				if( maxrgb == r && g >= b )
					H = 60 * (g-b) / diff + 0;
				else if( maxrgb == r && g < b )
					H = 60 * ((g-b) / diff) + 360;
				else if( maxrgb == g )
					H = 60 * ((b-r) / diff) + 120;
				else // if( maxrgb == b )
					H = 60 * ((r-g) / diff) + 240;

				// S = (maxrgb-minrgb)/255 = (maxrgb-minrgb)*(1/255);
				float S = float(maxrgb-minrgb) * div255;

				// h and s will be euclidean coordinates, so they can be negative as well as positive.
				// The length of (h,s) should be S.
				dest->pix32f(x,y,0) = Bully::cosf[H] * S;
				dest->pix32f(x,y,1) = Bully::sinf[H] * S;
				// L = ((maxrgb+minrgb)/2)/255 = (maxrgb+minrgb)*(1/510);
				dest->pix32f(x,y,2) = float(maxrgb + minrgb) * div2div255;
			}
		}
*/
	PROFILE_END_FUNC;
	return dest;
}
Bully::Buffer* Bully::Buffer::rgb2hsL_slow(Buffer *dest){
	assert( w() == dest->w() );
	assert( h() == dest->h() );
	assert( channels() == 3 && dest->channels() == 3 );
	assert( depth() == DEPTH_8U );
	assert( dest->depth() == DEPTH_32F );
	PROFILE_BEGIN_FUNC;
	int width = w();
	int height = h();

	for( int y = 0; y < height; y++ ){
		for( int x = 0; x < width; x++ ){
			float r = float(pix8u(x,y,0)) / 255.0f;
			float g = float(pix8u(x,y,1)) / 255.0f;
			float b = float(pix8u(x,y,2)) / 255.0f;
			float maxrgb = max(r,max(g,b));
			float minrgb = min(r,min(g,b));

			float H = 0.0f;
			if( maxrgb == r && g >= b )
				H = 60.0f * (g-b) / (maxrgb-minrgb) + 0.0f;
			else if( maxrgb == r && g < b )
				H = 60.0f * (g-b) / (maxrgb-minrgb) + 360.0f;
			else if( maxrgb == g )
				H = 60.0f * (b-r) / (maxrgb-minrgb) + 120.0f;
			else if( maxrgb == b )
				H = 60.0f * (r-g) / (maxrgb-minrgb) + 240.0f;

			float S = maxrgb - minrgb;
			float L = (maxrgb + minrgb) * 0.5f;

			// h and s will be euclidean coordinates, so they can be negative as well as positive.
			// The length of (h,s) should be S.
			float h = ::cosf(H*Bully::PIf/180.0f) * S;
			float s = ::sinf(H*Bully::PIf/180.0f) * S;

			dest->pix32f(x,y,0) = h;
			dest->pix32f(x,y,1) = s;
			dest->pix32f(x,y,2) = L;
		}
	}
	PROFILE_END_FUNC;
	return dest;
}

// Logical and arithmetic operations -----------------------------------------
Bully::Buffer* Bully::Buffer::setZero(){
	cvSetZero(_iplImage);
	return this;
}
Bully::Buffer* Bully::Buffer::set(uchar val, Buffer *mask){
	if( mask )
		cvSet(_iplImage, cvScalar(val), mask->iplImage());
	else
		cvSet(_iplImage, cvScalar(val));
	return this;
}
Bully::Buffer* Bully::Buffer::not_( Buffer *dest ){
	if( !dest ) dest = this;
	assert( depth() == dest->depth() );
	cvNot( _iplImage, dest->iplImage() );
	return dest;
}
Bully::Buffer* Bully::Buffer::and_( Buffer * mask, Buffer *dest ){
	if( !dest ) dest = this;
	assert( depth() == mask->depth() );
	assert( depth() == dest->depth() );
	cvAnd( _iplImage, mask->iplImage(), _iplImage );
	return dest;
}
Bully::Buffer* Bully::Buffer::or_( Buffer * mask, Buffer *dest ){
	if( !dest ) dest = this;
	assert( depth() == mask->depth() );
	assert( depth() == dest->depth() );
	cvOr( _iplImage, mask->iplImage(), _iplImage );
	return dest;
}
Bully::Buffer* Bully::Buffer::add( Buffer *src, Buffer *dest ){
	if( dest == 0 ) dest = this;
	assert( depth() == src->depth() );
	assert( depth() == dest->depth() );
	assert( channels() == src->channels() );
	cvAdd( _iplImage, src->iplImage(), dest->iplImage() );
	return dest;
}
Bully::Buffer* Bully::Buffer::sub( Buffer *src, Buffer *dest ){
	if( dest == 0 ) dest = this;
	assert( depth() == src->depth() );
	assert( depth() == dest->depth() );
	assert( channels() == src->channels() );
	cvAbsDiff( _iplImage, src->iplImage(), dest->iplImage() );
	return dest;
}
Bully::Buffer* Bully::Buffer::lessThan( Buffer *src, Buffer *dest ){
	if( dest == 0 ) dest = this;
	assert( depth() == src->depth() );
	assert( depth() == dest->depth() );
	assert( channels() == src->channels() && channels() == 1 );
	cvCmp( iplImage(), src->iplImage(), dest->iplImage(), CV_CMP_LT );
	return dest;
}

// Image processing  ---------------------------------------------------------
void Bully::Buffer::threshold( uchar t, Buffer *dest ){
	if( dest == 0 ) dest = this;

#if 0
	assert(0); // Something !#"!" strange happened here.
	cvThreshold( _iplImage, dest->iplImage(), t, 255, CV_THRESH_BINARY );
#elif 1
	cvThreshold( _iplImage, dest->iplImage(), t, 255, CV_THRESH_TOZERO );
	cvSet( dest->iplImage(), cvScalar(255), dest->iplImage() );
#elif 0
	uchar * data = (unsigned char *)(dest->iplImage()->imageData);
	for( int i = 0; i < dest->iplImage()->imageSize; i++ ){
		if( data[i] < t ) data[i] = 0;
		else data[i] = 0xff;
	}
#endif
}
void Bully::Buffer::adaptiveThreshold( int n, int c, Buffer *dest ){
	PROFILE_BEGIN_FUNC;
	// Cannot operate in-place.
	assert(dest && _name.compare(dest->name()) != 0);
	// The size of the neighborhood must be odd.
	assert( n%2 == 1 );
	cvAdaptiveThreshold( _iplImage, dest->iplImage(), 255.0,
                          CV_ADAPTIVE_THRESH_MEAN_C,
                          CV_THRESH_BINARY,
                          n, c );
	PROFILE_END_FUNC;
}
void Bully::Buffer::median( int n, Buffer *dest ){
	PROFILE_BEGIN_FUNC;
	// Cannot operate in-place.
	assert(dest && _name.compare(dest->name()) != 0);
	// The size of the neighborhood must be odd.
	assert( n%2 == 1 );
	cvSmooth( _iplImage, dest->iplImage(), CV_MEDIAN, n );
	PROFILE_END_FUNC;
}
void Bully::Buffer::blur( int n, Buffer *dest ){
	PROFILE_BEGIN_FUNC;
	if( dest == 0 ) dest = this;
	cvSmooth( _iplImage, dest->iplImage(), CV_BLUR, n );
	PROFILE_END_FUNC;
}
void Bully::Buffer::erode( _IplConvKernel * se, int it, Buffer *dest ){
	PROFILE_BEGIN_FUNC;
	if( dest == 0 ) dest = this;
	cvErode( _iplImage, dest->iplImage(), se, it );
	PROFILE_END_FUNC;
}
void Bully::Buffer::dilate( _IplConvKernel * se, int it, Buffer *dest ){
	PROFILE_BEGIN_FUNC;
	if( dest == 0 ) dest = this;
	cvDilate( _iplImage, dest->iplImage(), se, it );
	PROFILE_END_FUNC;
}

// Drawing  ------------------------------------------------------------------
void Bully::Buffer::line( int x1, int y1, int x2, int y2, uchar r, uchar g, uchar b ){
	cvLine( _iplImage, cvPoint(x1,y1), cvPoint(x2,y2), CV_RGB( r, g, b ) );
}
void Bully::Buffer::line( Vector2i &p1, Vector2i &p2, unsigned char r, unsigned char g, unsigned char b ){
	cvLine( _iplImage, cvPoint(p1.x,p1.y), cvPoint(p2.x,p2.y), CV_RGB( r, g, b ) );
}
void Bully::Buffer::circle( Vector2i &c, int rad, unsigned char r, unsigned char g, unsigned char b ){
	cvCircle( _iplImage, cvPoint(c.x,c.y), rad, CV_RGB(r,g,b) );
}
void Bully::Buffer::rect( int x1, int y1, int w, int h, uchar r, uchar g, uchar b ){
	cvRectangle( _iplImage, cvPoint(x1,y1), cvPoint(x1+w-1,y1+h-1), CV_RGB(r,g,b) );
}
void Bully::Buffer::rect( Rect<int> &rect, uchar r, uchar g, uchar b ){
	cvRectangle( _iplImage, cvPoint(rect.x,rect.y), cvPoint(rect.x+rect.w-1,rect.y+rect.h-1), CV_RGB(r,g,b) );
}
void Bully::Buffer::fillRect( int x1, int y1, int w, int h, uchar r, uchar g, uchar b ){
	cvRectangle( _iplImage, cvPoint(x1,y1), cvPoint(x1+w-1,y1+h-1), CV_RGB(r,g,b), CV_FILLED );
}
void Bully::Buffer::poly( vector<Vector2i> const &poly, unsigned char r, unsigned char g, unsigned char b ){
	int p = 0;
	for( ; p < int(poly.size())-1; p++ )
		line(poly[p].x, poly[p].y, poly[p+1].x, poly[p+1].y, r, g, b);
	if( p > 0 )
		line(poly[0].x, poly[0].y, poly[p].x, poly[p].y, r, g, b);
}
void Bully::Buffer::fillPoly( vector<Vector2i> const &poly, unsigned char r, unsigned char g, unsigned char b ){
	int pointCount = poly.size();
	CvPoint *pts = new CvPoint[pointCount];
	for( int p = 0; p < pointCount; p++){
		pts[p].x = poly[p].x;
		pts[p].y = poly[p].y;
	}
	cvFillConvexPoly( _iplImage, pts, pointCount, cvScalar(r,g,b));
	delete[] pts;
}
void Bully::Buffer::drawContour( std::vector<Vector2i> const &contour, uchar r, uchar g, uchar b ){
	if( channels() == 1 ){
		for( vector<Vector2i>::const_iterator it = contour.begin(); it != contour.end(); it++ )
			pix8u(it->x, it->y) = r;
	}else if( channels() == 3 ){
		for( vector<Vector2i>::const_iterator it = contour.begin(); it != contour.end(); it++ ){
			pix8u(it->x, it->y, 0) = b;
			pix8u(it->x, it->y, 1) = g;
			pix8u(it->x, it->y, 2) = r;
		}
	}else{
		assert(0);
	}
}
int Bully::Buffer::floodFill( int x, int y, uchar value, Recti *bbox ){
	CvConnectedComp comp;
	cvFloodFill( _iplImage, cvPoint(x,y), cvScalar(value), cvScalarAll(0), cvScalarAll(0), &comp, 8 );
	if( bbox ){
		bbox->x = comp.rect.x;
		bbox->y = comp.rect.y;
		bbox->w = comp.rect.width;
		bbox->h = comp.rect.height;
	}
	return comp.area;
}

// Support and debug  --------------------------------------------------------
string Bully::Buffer::writeToFile(){
	return writeToFile(_name + string(".bg"));
}
string Bully::Buffer::writeToFile(std::string& path){
	assert( depth() == DEPTH_8U || depth() == DEPTH_32F );

	BGFile bgFile;
	bgFile.width = _width;
	bgFile.height = _height;
	bgFile.channels = _channels;

	if( depth() == DEPTH_8U ){
		bgFile.depth = _depth;
		bgFile.data = _data;
		writeBGFile( path.c_str(), bgFile );
	}else if( depth() == DEPTH_32F ){
		unsigned char *idata = new unsigned char[_width*_height*_channels];

		// Find the maximum channel value, and use that to normalize the colors to the range [0, 255].
		double maxCol = 0.0;
		for( int i = 1; i <= _channels; i++ ){
			cvSetImageCOI( iplImage(), i );
			double mintmp, maxtmp;
			cvMinMaxLoc( iplImage(), &mintmp, &maxtmp );
			maxCol = max(maxCol, maxtmp);
		}
		cvSetImageCOI( iplImage(), 0 );
		cvResetImageROI( iplImage() );

		float normFactor = 255.0f / float(maxCol);

		for( int y = 0; y < _height; y++ ){
			for( int x = 0; x < _width; x++ ){
				for( int c = 0; c < _channels; c++ ){
					float f = _fdata[y*_width*_channels + x*_channels + c] * normFactor;
					idata[y*_width*_channels + x*_channels + c] = (unsigned char)(f);
				}
			}
		}

		bgFile.depth = DEPTH_8U;
		bgFile.data = idata;
		writeBGFile( path.c_str(), bgFile );

		delete[] idata;
	}
	return path;
}
void Bully::Buffer::show(){
	string path = writeToFile();
#ifdef WIN32
	string cmd = string("start /B bgtool ") + path;
	system(cmd.c_str());
#else
	// Not implemented!
	assert(0);
#endif
}
std::string Bully::Buffer::printInfo(){
	stringstream ss;
	ss	<< "Buffer name: " << _name << endl
		<< "       w x h: " << _width << " x " << _height << endl
		<< "       channels: " << _channels << endl
		<< "       depth: ";
	if( _depth == DEPTH_8U )
		ss << "uchar";
	else
		ss << "float";
	return ss.str();
}

// Text io-routines ----------------------------------------------------------
#include "bullyfont.h"

void Bully::Buffer::invertPixel(int x, int y){
	assert(_depth == DEPTH_8U);
	if( x < 0 || x >= _width || y < 0 || y >= _height ) return;
	if( _channels == 1 ){
		uchar& c = pix8u(x,y);
		c = 255 - c;
	}else{
		assert(_channels == 3);
		uchar &r = pix8u(x,y,0);
		uchar &g = pix8u(x,y,1);
		uchar &b = pix8u(x,y,2);
		r = 255 - r;
		g = 255 - g;
		b = 255 - b;
	}
}
void Bully::Buffer::setPixel(int x, int y, uchar r, uchar g, uchar b){
	assert(_depth == DEPTH_8U);
	if( x < 0 || x >= _width || y < 0 || y >= _height ) return;
	if( _channels == 1 ){
		pix8u(x,y) = r;
	}else{
		assert(_channels == 3);
		pix8u(x,y,0) = r;
		pix8u(x,y,1) = g;
		pix8u(x,y,2) = b;
	}
}
void Bully::Buffer::print( int x, int y, char c ){
	static char tmpbuf[2];
	tmpbuf[0] = c;
	tmpbuf[1] = '\0';
	print(x,y, tmpbuf);
}
void Bully::Buffer::print( int x, int y, const char * text ){
	static bool initialized = false;
	static int xpmPixelOffset = 0;
	if( !initialized ){
		stringstream info;
		info << font[0];
		int columns, rows, colors, charsperpixel;
		info >> columns;
		info >> rows;
		info >> colors;
		info >> charsperpixel;
		xpmPixelOffset = colors + 2;
		initialized = true;
	}

	for( unsigned int i = 0; i < strlen(text); i++ ){
		int ascii = (int)(text[i]) - 32;
		int xpmRow = xpmPixelOffset + (ascii/16) * 16;
		int xpmCol = (ascii % 16) * 16;

		int bufY = y;
		for( int xpmY = xpmRow; xpmY < xpmRow + 16; xpmY++, bufY++ ){
			int bufX = x + i*16;
			for( int xpmX = xpmCol; xpmX < xpmCol + 16; xpmX++, bufX++ ){
				if( font[xpmY][xpmX] != ' ' ){
					setPixel(bufX, bufY, 255,0,255);
//					invertPixel(bufX, bufY);
				}
			}
		}
	}
}
