#pragma once

#include <string>
#include <assert.h>
#include <vector>
#include "bullytypes.h"

struct _IplImage;
struct _IplConvKernel;

namespace Bully {

class Buffer {
public:
	Buffer( const char * name, int w, int h, int c, PixelDepth d );
	~Buffer();

	void copy( Buffer * srcBuffer, Buffer * mask=0 );
	void copy( unsigned char * srcData );
	void copy( int srcw, int srch, int srcbpp, unsigned char * srcData );

	// Color conversion
	Buffer* gray2rgb(Buffer *dest);
	Buffer* rgb2gray(Buffer *dest);
	Buffer* rgb2gray_erik(Buffer *dest);
	Buffer* rgb2hsv(Buffer *dest);
	Buffer* rgb2hsL(Buffer *dest);
	Buffer* rgb2hsL_slow(Buffer *dest);

	// Logical and arithmetic operations
	Buffer* setZero();
	Buffer* set( unsigned char val, Buffer *mask=0 );
	Buffer* not_( Buffer *dest=0 );
	Buffer* and_( Buffer *mask, Buffer *dest=0 );
	Buffer* or_( Buffer *mask, Buffer *dest=0 );
	Buffer* add( Buffer *src, Buffer *dest=0 );
	Buffer* sub( Buffer *src, Buffer *dest=0 );
	Buffer* lessThan( Buffer *src, Buffer *dest=0 );

	// Image Processing
	void threshold( unsigned char t, Buffer *dest=0 );
	void blur( int n, Buffer *dest=0 );
	void adaptiveThreshold( int n, int c, Buffer *dest );
	void median( int n, Buffer *dest );
	void erode( _IplConvKernel * se, int it, Buffer *dest=0 );
	void dilate( _IplConvKernel * se, int it, Buffer *dest=0 );

	// Drawing
	void print( int x, int y, const char * text );
	void print( int x, int y, char c );
	void line( int x1, int y1, int x2, int y2, unsigned char r, unsigned char g=0, unsigned char b=0 );
	void line( Vector2i &p1, Vector2i &p2, unsigned char r, unsigned char g=0, unsigned char b=0 );
	void circle( Vector2i &center, int radius, unsigned char r, unsigned char g=0, unsigned char b=0 );
	void rect( Rect<int> &rect, unsigned char r, unsigned char g=0, unsigned char b=0 );
	void rect( int x1, int y1, int w, int h, unsigned char r, unsigned char g=0, unsigned char b=0 );
	void fillRect( int x1, int y1, int w, int h, unsigned char r, unsigned char g=0, unsigned char b=0 );
	void poly( std::vector<Vector2i> const &poly, unsigned char r, unsigned char g=0, unsigned char b=0 );
	void fillPoly( std::vector<Vector2i> const &poly, unsigned char r, unsigned char g=0, unsigned char b=0 );
	void drawContour( std::vector<Vector2i> const &contour, unsigned char r, unsigned char g=0, unsigned char b=0 );
	int floodFill( int x, int y, unsigned char value, Recti *bbox=0 );

	// Support and debug
	std::string writeToFile();
	std::string writeToFile(std::string& path);
	void show();
	std::string printInfo();

	// Accessors
	inline std::string name(){ return _name; }
	inline void name(const char *newName){ _name = newName; }
	inline int w(){ return _width; }
	inline int h(){ return _height; }
	inline PixelDepth depth(){ return _depth; }
	inline int channels(){ return _channels; }
	inline int pitch(){ return _pitch; }
	inline int size(){ return _size; }
	inline unsigned char * data(){ return _data; }
	inline float * fdata(){ return _fdata; }
	inline _IplImage * iplImage(){ return _iplImage; }

	// Convenience methods.
	void invertPixel(int x, int y);
	void setPixel(int x, int y, unsigned char r,unsigned char g,unsigned char b);
	inline float & pix32f(int x, int y, int c){
		assert( x >= 0 && x < _width &&
				y >= 0 && y < _height &&
				c >= 0 && c < _channels &&
				_depth == DEPTH_32F );
		return _fdata[y*_width*_channels + x*_channels + c];
	}
	inline unsigned char & pix8u(int x, int y, int c){
		assert( x >= 0 && x < _width &&
				y >= 0 && y < _height &&
				c >= 0 && c < _channels &&
				_depth == DEPTH_8U );
		return _data[y*_pitch + x*_channels + c];
	}
	inline unsigned char & pix8u(int x, int y){
		assert( x >= 0 && x < _width && y >= 0 && y < _height &&
				_channels == 1 && _depth == DEPTH_8U );
		return _data[y*_width+x];
	}
	inline unsigned char & pix8u(int i){
		assert( i >= 0 && i < _width*_height &&
				_channels == 1 && _depth == DEPTH_8U );
		return _data[i];
	}
	inline unsigned char & operator()(int x, int y){ return pix8u(x, y); }
	inline unsigned char & operator()(int i){ return pix8u(i); }

	inline bool inside( int x, int y ) const {
		return (x >= 0 && x < _width && y >= 0 && y < _height);
	}

protected:
	void init();
	void init( const char * name, int w, int h, int c, PixelDepth d );

	_IplImage * _iplImage;

	int _width;
	int _height;
	PixelDepth _depth;
	int _channels;
	int _pitch;
	int _size;
	unsigned char * _data;
	float * _fdata;
	std::string _name;
};

}; // namespace Bully
