#include "bullymoduleprep.h"
#include "bullybuffermanager.h"
#include "bullybuffer.h"
#include "bullyparams.h"
#include "bullybloblist.h"
#include <cv.h>
#include <profiling.h>
#include <string>
#include <sstream>
using namespace std;

namespace Bully {

ModulePreprocess::ModulePreprocess()
	: Module()
{
}
ModulePreprocess::ModulePreprocess(BufferManager &buffer, Params &params)
	: Module(buffer, params)
{
	blobList = new BlobList(256);

	// Create morphological structuring element.
	int s = (int)this->params->get("prepStructElementSize");
	int o = s/2;
	structElement = cvCreateStructuringElementEx( s, s, o, o, CV_SHAPE_RECT );

	// Create look-up table for pseudo-coloring regions.
	pseudoColors[0][0] = 0;
	pseudoColors[0][1] = 0;
	pseudoColors[0][2] = 0;
	CvRNG rngState = cvRNG(0xffffffff);
	for( int i = 1; i < 256; i++ ){
		int b = cvRandInt(&rngState)%256;
		int g = cvRandInt(&rngState)%256;
		int r = cvRandInt(&rngState)%256;
		if( b + g + r < 300 ){
			// Make sure the color is reasonably bright.
			int smallest = min(r,min(g,b));
			if( r == smallest )			r = cvRandInt(&rngState)%128 + 128;
			else if( g == smallest )	g = cvRandInt(&rngState)%128 + 128;
			else						b = cvRandInt(&rngState)%128 + 128;
		}
		pseudoColors[i][0] = (uchar)b;
		pseudoColors[i][1] = (uchar)g;
		pseudoColors[i][2] = (uchar)r;
	}
}
Bully::ModulePreprocess::~ModulePreprocess(){
	delete blobList; blobList = 0;
	cvReleaseStructuringElement( &structElement );
}

void Bully::ModulePreprocess::removeSmallBlobs(const char *src, const char *dst){
	PROFILE_BEGIN_FUNC;
	Buffer *srcbuffer = buffer->get(src);
	Buffer *dstbuffer = buffer->get(dst);
	int w = srcbuffer->w();
	int h = srcbuffer->h();

	dstbuffer->copy(srcbuffer);
	Buffer *tmp = buffer->createBuffer(w,h,1,DEPTH_8U,"tmp");
	tmp->copy(srcbuffer);

	double minInterestArea = params->get("prepRegionAreaLower");
	CvConnectedComp comp;

	for( int y = 0; y < h; y++ ){
		for( int x = 0; x <w; x++ ){

			if( tmp->pix8u(x,y) == 0xFF ){

				cvFloodFill( tmp->iplImage(), cvPoint(x,y), cvScalar(0),
							 cvScalarAll(0), cvScalarAll(0), &comp, 8 );

				if( comp.area < minInterestArea ){
					// It is too small to be of interest, so remove it.
					cvFloodFill( dstbuffer->iplImage(), cvPoint(x,y), cvScalar(0),
								cvScalarAll(0), cvScalarAll(0), 0, 8 );
				}
			}
		}
	}
	PROFILE_END_FUNC;
}

void Bully::ModulePreprocess::labelBlobs(const char *src, const char *dst){
	PROFILE_BEGIN_FUNC;
	// Before we change "preprocessed", copy it into "foreground".
	Buffer *srcbuffer = buffer->get(src);
	Buffer *dstbuffer = buffer->get(dst);
	dstbuffer->copy(srcbuffer);

	// Background has value/label 0.
	// Unlabelled foreground has value 0xFF
	int labelCount = 1;
	CvConnectedComp comp;

	for( int y = 0; y < dstbuffer->h(); y++ ){
		for( int x = 0; x < dstbuffer->w(); x++ ){

			if( dstbuffer->pix8u(x,y) == 0xFF ){

				assert(labelCount < 255);

				cvFloodFill( dstbuffer->iplImage(), cvPoint(x,y), cvScalar(labelCount),
							 cvScalarAll(0), cvScalarAll(0), &comp, 8 );

				Blob b;
				b.id(labelCount);
				b.frameFirstDetected(_frameCount);
				b.frameName(src);
				b.area = (int)comp.area;
				b.bbox = Recti(comp.rect.x, comp.rect.y, comp.rect.width, comp.rect.height);
				b.col[0] = pseudoColors[labelCount][0];
				b.col[1] = pseudoColors[labelCount][1];
				b.col[2] = pseudoColors[labelCount][2];
				traceOuterContour(dst, x, y, b);
				b.centroid = findCentroid(dst, b);
				blobList->add(b);

				++labelCount;
			}
		}
	}
	PROFILE_END_FUNC;
}
void Bully::ModulePreprocess::pseudoColorBlobs(const char *src, const char *dst, bool allwhite){
	PROFILE_BEGIN_FUNC;

	Buffer *srcbuffer = buffer->get(src);
	Buffer *dstbuffer = buffer->get(dst);
	int w = srcbuffer->w();
	int h = srcbuffer->h();

	if( allwhite ){
		cvSet(dstbuffer->setZero()->iplImage(), cvScalar(255,255,255), srcbuffer->iplImage());
	}else{
		for( int y = 0; y < h; y++ ){
			for( int x = 0; x < w; x++ ){
				dstbuffer->pix8u(x,y,0) = pseudoColors[srcbuffer->pix8u(x,y)][0];
				dstbuffer->pix8u(x,y,1) = pseudoColors[srcbuffer->pix8u(x,y)][1];
				dstbuffer->pix8u(x,y,2) = pseudoColors[srcbuffer->pix8u(x,y)][2];
			}
		}
	}
	PROFILE_END_FUNC;
}
Bully::Vector2i Bully::ModulePreprocess::findCentroid(const char *src, Blob & b){
	PROFILE_BEGIN_FUNC;

	assert(b.area > 0);
	Buffer *srcbuffer = buffer->get(src);
	string centroidType = params->getString("prepPosition");

	Vector2i centroid;

	if( centroidType == "centroid" ){

		int xsum=0, ysum=0;
		unsigned char * pix = srcbuffer->data() + b.bbox.y*srcbuffer->pitch();

		for(int y = b.bbox.y; y < b.bbox.y + b.bbox.h; y++){
			for(int x = b.bbox.x; x < b.bbox.x + b.bbox.w; x++){
				if( pix[x] == b.id() ){
					xsum += x;
					ysum += y;
				}
			}
			pix += srcbuffer->pitch();
		}
		centroid = Vector2i(xsum/b.area, ysum/b.area);

	}else if( centroidType == "median" ){

		int * xcoords = new int[b.bbox.w];
		int * ycoords = new int[b.bbox.h];
		memset(xcoords, 0, b.bbox.w*sizeof(int));
		memset(ycoords, 0, b.bbox.h*sizeof(int));
		unsigned char * pix = srcbuffer->data() + b.bbox.y*srcbuffer->pitch();

		for(int y = 0; y < b.bbox.h; y++){
			for(int x = 0; x < b.bbox.w; x++){
				if( pix[b.bbox.x + x] == b.id() ){
					xcoords[x]++;
					ycoords[y]++;
				}
			}
			pix += srcbuffer->pitch();
		}

		int t = b.area / 2;
		int xmed=0, ymed=0;

		for( int sum = 0; ymed < b.bbox.h && sum < t; ymed++ )
			sum += ycoords[ymed];

		for( int sum = 0; xmed < b.bbox.w && sum < t; xmed++ )
			sum += xcoords[xmed];

		delete[] xcoords;
		delete[] ycoords;
		centroid = Vector2i(b.bbox.x + xmed, b.bbox.y + ymed);

	}else{ // if centroidType == "centerbbox"

		centroid = Vector2i(b.bbox.x + b.bbox.w / 2, b.bbox.y + b.bbox.h / 2);
	}

	PROFILE_END_FUNC;
	return centroid;
}
void Bully::ModulePreprocess::getOuterContour(const char *src, Recti &bbox, uchar label){
	Buffer *srcbuffer = buffer->get(src);
	Buffer *outerContour = buffer->createBuffer(srcbuffer,"outercontour")->setZero();
	int w = srcbuffer->w();
	int h = srcbuffer->h();

	for( int px = max(1,bbox.x-1); px < min(w-1,bbox.x+bbox.w+1); px++ ){
		for( int py = max(1,bbox.y-1); py < min(h-1,bbox.y+bbox.h+1); py++ ){
			if( srcbuffer->pix8u(px,py) == 0 ){
				if( srcbuffer->pix8u(px-1,py) == label ||
					srcbuffer->pix8u(px,py-1) == label ||
					srcbuffer->pix8u(px+1,py) == label ||
					srcbuffer->pix8u(px,py+1) == label )
				{
					outerContour->pix8u(px,py) = 255;
				}
			}
		}
	}
}
void Bully::ModulePreprocess::traceOuterContour(const char *src, int x, int y, Blob & b){
	uchar label = b.id();
	getOuterContour(src, b.bbox, label);

	Buffer *contour = buffer->get("outercontour");
	int w = contour->w();
	int h = contour->h();

	Vector2i p = Vector2i(x,y);
	b.contour.startPoint = p;

	for(;;){
		b.contour.pixels.push_back(p);
		contour->pix8u(p.x, p.y) = 127;

		if( p.y > 0	&& contour->pix8u(p.x,p.y-1) == 255 ){
			p.y--;
		}else if( p.x < w-1 && contour->pix8u(p.x+1,p.y) == 255 ){
			p.x++;
		}else if( p.y < h-1 && contour->pix8u(p.x,p.y+1) == 255 ){
			p.y++;
		}else if( p.x > 0 && contour->pix8u(p.x-1,p.y) == 255 ){
			p.x--;
		}else if( p.y > 0 && p.x < w-1 && contour->pix8u(p.x+1,p.y-1) == 255 ){
			p.x++;
			p.y--;
		}else if( p.y < h-1 && p.x < w-1 && contour->pix8u(p.x+1,p.y+1) == 255 ){
			p.x++;
			p.y++;
		}else if( p.y < h-1 && p.x > 0 && contour->pix8u(p.x-1,p.y+1) == 255 ){
			p.x--;
			p.y++;
		}else if( p.y > 0 && p.x > 0 && contour->pix8u(p.x-1,p.y-1) == 255 ){
			p.x--;
			p.y--;
		}else{
			break;
		}
	}

//	assert((p - b.contour.startPoint).abs() < 2.0);
}
inline bool Bully::ModulePreprocess::getneighbor(Vector2i &C, int S, int w, int h, Vector2i &N){
	bool retval = true;
	if( S >= 3 && S <= 5 )
		N.x = C.x - 1;
	else if( S != 2 && S != 6 )
		N.x = C.x + 1;
	if( S >= 1 && S <= 3 )
		N.y = C.y - 1;
	else if( S != 0 && S != 4 )
		N.y = C.y + 1;

	if( N.x < 0 || N.x >= w || N.y < 0 || N.y >= h )
		retval = false;
	return retval;
}
void Bully::ModulePreprocess::traceContour(const char *src, int x, int y, Blob & b){
	PROFILE_BEGIN_FUNC;
	b.contour.startPoint = Vector2i(x,y);
	Buffer *srcbuffer = buffer->get(src);
	int w = srcbuffer->w();
	int h = srcbuffer->h();

	Vector2i C = b.contour.startPoint;
	int S = 6;

	do {
		b.contour.pixels.push_back(C);
		for( int i = 0; i < 3; i++ ){
			Vector2i N;
			if( getneighbor(C,(S+7)%8, // (S-1)%8; Op. % does not handle neg. numbers
			w,h,N) && srcbuffer->pix8u(N.x,N.y) == b.id() ){
				C = N;
				S = (S+6)%8; // (S-2)%8;  Op. % does not handle neg. numbers
				break;
			}else if( getneighbor(C,S,w,h,N) && srcbuffer->pix8u(N.x,N.y) == b.id() ){
				C = N;
				break;
			}else if( getneighbor(C,(S+1)%8,w,h,N) && srcbuffer->pix8u(N.x,N.y) == b.id() ){
				C = N;
				break;
			}else{
				S = (S+2)%8;
			}
		}

		b.contour.chaincode.push_back(S);
	}while( C != b.contour.startPoint );

	b.contour.chaincode.pop_back();
	PROFILE_END_FUNC;
}
void Bully::ModulePreprocess::buildHistograms(){
	PROFILE_BEGIN_FUNC;

	Buffer* input = buffer->get("input_lum");
	Buffer* labeled = buffer->get("preprocessed");

	for( int i = 0; i < blobList->count(); i++ ){
		Blob & b = *(blobList->get(i));

		double total = .0;
		for( int y = b.bbox.y; y < b.bbox.y + b.bbox.h; y++ ){
			for( int x = b.bbox.x; x < b.bbox.x + b.bbox.w; x++ ){
				if( labeled->pix8u(x,y) == b.id() ){
					b.hist[ input->pix8u(x,y) ]++;
					total++;
				}
			}
		}

		for( int i = 0; i < 256; i++ )
			b.hist[i] /= total;
	}
	PROFILE_END_FUNC;
}
void ModulePreprocess::renderROIBuffer(const char *dst){
	if( (int)params->valueCount("roi") == 0 ){
		// Make the whole buffer "interesting"
		buffer->get(dst)->set(255);
	}else{
		for( int i = 0; i < (int)params->valueCount("roi"); i++ ){
			vector<Vector2i> roi;
			params->getVector(params->getString("roi", i), roi);
			buffer->get(dst)->fillPoly(roi, 255,255,255);
		}
	}

	for( int i = 0; i < (int)params->valueCount("roni"); i++ ){
		vector<Vector2i> roni;
		params->getVector(params->getString("roni", i), roni);
		buffer->get(dst)->fillPoly(roni, 0,0,0);
	}
}
void ModulePreprocess::drawDebugOutput(){
	if( (int)params->get("drawInput") == 1 )
		buffer->get("output")->copy(buffer->get("input_rgb"));

	if( (int)params->get("drawROI") == 1 ){
		for( int i = 0; i < (int)params->valueCount("roi"); i++ ){
			vector<Vector2i> roi;
			params->getVector(params->getString("roi", i), roi);
			buffer->get("output")->poly(roi, 255,0,0);
		}
		for( int i = 0; i < (int)params->valueCount("roni"); i++ ){
			vector<Vector2i> roni;
			params->getVector(params->getString("roni", i), roni);
			buffer->get("output")->poly(roni, 0,0,255);
		}
	}

	if( (int)params->get("drawShadow") == 1 ){
		if( buffer->exists("shadow") )
			cvSet( buffer->get("output")->iplImage(), cvScalar(255,255,255), buffer->get("shadow")->iplImage() );
	}

	if( (int)params->get("drawBlobContour") == 1 ){
		for( int i = 0; i < blobList->count(); i++ ){
			Blob & b = *(blobList->get(i));
			buffer->get("output")->drawContour(b.contour.pixels, b.col[0], b.col[1], b.col[2]);
		}
	}

	if( (int)params->get("drawBlobRect") == 1 ){
		for( int i = 0; i < blobList->count(); i++ ){
			Blob & b = *(blobList->get(i));
			buffer->get("output")->rect( b.bbox.x, b.bbox.y, b.bbox.w, b.bbox.h, 0xFF, 0x00, 0xFF );
		}
	}
}

} // namespace Bully
