#include "bullytypes.h"
#include "bullymoduleprep_cncc_hsl.h"
#include "bullyparams.h"
#include "bullybuffer.h"
#include "bullybuffermanager.h"
#include "bullybloblist.h"
#include <time.h>
#include <cv.h>
#include <algorithm>
#include <profiling.h>
#include "eriksul.h"
using namespace std;

#pragma intrinsic( sqrt, sin, cos, abs, fabs )

Bully::ModulePreprocess_CNCC_hsL::ModulePreprocess_CNCC_hsL( BufferManager &buffer, Params &params )
: ModulePreprocess(buffer, params) {
}
Bully::ModulePreprocess_CNCC_hsL::~ModulePreprocess_CNCC_hsL(){
}

void Bully::ModulePreprocess_CNCC_hsL::handleInput( bool down, char ascii ){
	if( ascii == 's' ){
		buffer->get("pseudo")->show();
		buffer->get("shadow")->show();
//		buffer->get("input_hsL")->show();
//		buffer->get("background_hsL")->show();
	}
}

void Bully::ModulePreprocess_CNCC_hsL::process(){
	PROFILE_BEGIN_FUNC;
	++_frameCount;

	// ---------------------------------------------------------------
	// NB: this method uses a lot of memory in the form of unnecessary
	// buffers. This is purely for debugging purposes.
	// ---------------------------------------------------------------

	// Clear results from last frame.
	blobList->clear();

	// Create buffers.
	int w = buffer->get("background_rgb")->w();
	int h = buffer->get("background_rgb")->h();
	if( !buffer->exists("background_lum") ){
		buffer->createBuffer( w, h, 1, DEPTH_8U, "background_lum" );
		buffer->get("background_rgb")->rgb2gray(buffer->get("background_lum"));
	}
	buffer->createBuffer( w, h, 1, DEPTH_8U, "tmp" );

	// Convert to greyscale
	buffer->createBuffer( w, h, 1, DEPTH_8U, "input_lum" );
	buffer->get("input_rgb")->rgb2gray(buffer->get("input_lum"));

	// Subtract background
	buffer->createBuffer( w, h, 1, DEPTH_8U, "difference" );
	buffer->get("input_lum")->sub( buffer->get("background_lum"), buffer->get("difference"));

	// Threshold difference image
	buffer->createBuffer( w, h, 1, DEPTH_8U, "tdifference" );
	buffer->get("difference")->threshold((int)params->get("prepForegroundDiffLower"), buffer->get("tdifference"));

	// Mask out anything not in a region of interest.
	buffer->createBuffer(w, h, 1, DEPTH_8U, "roi");
	renderROIBuffer("roi");
	buffer->get("tdifference")->and_(buffer->get("roi"));

	// Erode once to remove one-pixel-width noise
	buffer->createBuffer( w, h, 1, DEPTH_8U, "eroded" );
	buffer->get("tdifference")->erode(structElement, 1, buffer->get("eroded"));

	// Remove blobs that are too small
	buffer->createBuffer( w, h, 1, DEPTH_8U, "smallblobsremoved" );
	removeSmallBlobs("eroded", "smallblobsremoved");

	// Find the pixels that are candidates for shadow pixels
	buffer->createBuffer( w, h, 1, DEPTH_8U, "shadowcandidates" );
	buffer->get("input_lum")->lessThan(buffer->get("background_lum"), buffer->get("shadowcandidates"));
	buffer->get("shadowcandidates")->and_(buffer->get("smallblobsremoved"));

	// Detect and remove shadows
	buffer->createBuffer( w, h, 1, DEPTH_8U, "shadow" )->setZero();
	findShadowPixels("shadowcandidates", "shadow");

	buffer->createBuffer( w, h, 1, DEPTH_8U, "foreground" );
	buffer->get("foreground")->copy(buffer->get("smallblobsremoved"));
	buffer->get("shadow")->not_(buffer->get("tmp"));
	buffer->get("foreground")->and_(buffer->get("tmp"));

	buffer->get("foreground")->dilate(structElement, 2);
	buffer->get("foreground")->erode(structElement, 1);
	// Remove any shadowcandidates that were rejected.
	buffer->get("foreground")->not_(buffer->get("tmp"));
	buffer->get("shadow")->and_(buffer->get("tmp"));

	// Label the blobs and add to the blobList.
	buffer->createBuffer( w, h, 1, DEPTH_8U, "labeled" );
	labelBlobs("foreground", "labeled");

	// Pseudo-color the blobs
	buffer->createBuffer( w, h, 3, DEPTH_8U, "pseudo" );
	pseudoColorBlobs("labeled", "pseudo");

	drawDebugOutput();
	PROFILE_END_FUNC;
}
void Bully::ModulePreprocess_CNCC_hsL::findShadowPixels(const char *src, const char *dst){
	// Remove shadows using Grest, Frahm and Koch: "A Color Similarity Measure
	// for Robust Shadow Removal in Real-Time".
	PROFILE_BEGIN_FUNC;
	Buffer* srcbuffer = buffer->get(src);
	Buffer* dstbuffer = buffer->get(dst)->setZero();
	int w = srcbuffer->w();
	int h = srcbuffer->h();

	if( !buffer->exists("background_hsL") ){
		// This happens only the first iteration.
		buffer->createBuffer( w, h, 3, DEPTH_32F, "background_hsL" );
		buffer->get("background_rgb")->rgb2hsL(buffer->get("background_hsL"));
	}
	Buffer* b_hsL = buffer->get("background_hsL");
	Buffer* i_hsL = buffer->createBuffer( w, h, 3, DEPTH_32F, "input_hsL" );
	buffer->get("input_rgb")->rgb2hsL(i_hsL);

	int NW = (int)params->get("prepShadowCNCCWidth");
	assert(NW >= 7 && NW % 2 == 1);
	int N = NW/2;
	float MN = float(NW*NW);
	float divMN = 1.0f / MN;
	float shadowThreshold = (float)params->get("prepShadowCNCCLower");

	for( int y = N; y < h - N; y++ ){
		for( int x = N; x < w - N; x++ ){
			if( srcbuffer->pix8u(x,y) == 0 ){
				// Not a candidate (not moving OR brighter than background).
				continue;
			}

			// Candidate shadow pixel.
			float lumi=0.0f, lumb=0.0f, sumi=0.0f, sumb=0.0f, sumib=0.0f;
			for( int j = y - N; j < y + N; j++ ){
				for( int i = x - N; i < x + N; i++ ){
					lumi += i_hsL->pix32f(i,j,2);
					lumb += b_hsL->pix32f(i,j,2);

					sumi +=		max(.0f,i_hsL->pix32f(i,j,0) * i_hsL->pix32f(i,j,0) +
										i_hsL->pix32f(i,j,1) * i_hsL->pix32f(i,j,1)) +
								i_hsL->pix32f(i,j,2) * i_hsL->pix32f(i,j,2);

					sumb +=		max(.0f,b_hsL->pix32f(i,j,0) * b_hsL->pix32f(i,j,0) +
										b_hsL->pix32f(i,j,1) * b_hsL->pix32f(i,j,1)) +
								b_hsL->pix32f(i,j,2) * b_hsL->pix32f(i,j,2);

					sumib +=	max(.0f,i_hsL->pix32f(i,j,0) * b_hsL->pix32f(i,j,0) +
										i_hsL->pix32f(i,j,1) * b_hsL->pix32f(i,j,1)) +
								i_hsL->pix32f(i,j,2) * b_hsL->pix32f(i,j,2);
				}
			}
			lumi *= divMN;
			lumb *= divMN;
			float vari = sumi - MN*lumi*lumi;
			float varb = sumb - MN*lumb*lumb;
			// cncc = (sumib-MN*lumi*lumb) / sqrt(vari*varb)
			float cncc = (sumib - MN*lumi*lumb) * cvInvSqrt(vari*varb);

			if( cncc >= shadowThreshold ){
				// Probably a shadow pixel.
				dstbuffer->pix8u(x,y) = 0xFF;
			}
		}
	}
	PROFILE_END_FUNC;
}
void Bully::ModulePreprocess_CNCC_hsL::drawDebugOutput(){
	ModulePreprocess::drawDebugOutput();
}
