#include "bullytypes.h"
#include "bullymoduleprep_hsv.h"
#include "bullyparams.h"
#include "bullybuffer.h"
#include "bullybuffermanager.h"
#include "bullybloblist.h"
#include <time.h>
#include <cv.h>
#include <algorithm>
#include <profiling.h>
#include "eriksul.h"
using namespace std;

#pragma intrinsic( sqrt, sin, cos, abs, fabs )

Bully::ModulePreprocess_HSV::ModulePreprocess_HSV( BufferManager &buffer, Params &params )
: ModulePreprocess(buffer, params) {
}
Bully::ModulePreprocess_HSV::~ModulePreprocess_HSV(){
}

void Bully::ModulePreprocess_HSV::handleInput( bool down, char ascii ){
	if( ascii == 's' ){
		buffer->get("foreground")->show();
		buffer->get("shadow")->show();
		buffer->get("shadowcandidates")->show();
	}
}

// Input: "background" and "input": 8-bit images.
// Output: "preprocessed": binary image (0 and 255 are the only values).
void Bully::ModulePreprocess_HSV::process(){
	PROFILE_BEGIN_FUNC;
	++_frameCount;

	// Clear results from last frame.
	blobList->clear();

	// Create buffers and convert to grayscale (luminance).
	int w = buffer->get("background_rgb")->w();
	int h = buffer->get("background_rgb")->h();
	if( !buffer->exists("background_lum") ){
		buffer->createBuffer( w, h, 1, DEPTH_8U, "background_lum" );
		buffer->get("background_rgb")->rgb2gray(buffer->get("background_lum"));
	}
	buffer->createBuffer( w, h, 1, DEPTH_8U, "input_lum" );
	buffer->get("input_rgb")->rgb2gray(buffer->get("input_lum"));

	// Subtract background luminance from input luminance.
	buffer->createBuffer( w, h, 1, DEPTH_8U, "difference" );
	buffer->get("input_lum")->sub(buffer->get("background_lum"), buffer->get("difference"));

	// Threshold difference image.
	buffer->createBuffer( w, h, 1, DEPTH_8U, "tdifference" );
	buffer->get("difference")->threshold((int)params->get("prepForegroundDiffLower"), buffer->get("tdifference"));

	// Mask out anything not in a region of interest.
	buffer->createBuffer(w, h, 1, DEPTH_8U, "roi");
	renderROIBuffer("roi");
	buffer->get("tdifference")->and_(buffer->get("roi"));

	// Remove noise by applying erosion and removing any remaining regions too small.
	buffer->createBuffer( w, h, 1, DEPTH_8U, "eroded" );
	buffer->get("tdifference")->erode(structElement, 1, buffer->get("eroded"));

	// Remove blobs that are too small.
	buffer->createBuffer( w, h, 1, DEPTH_8U, "smallblobsremoved" );
	removeSmallBlobs("eroded", "smallblobsremoved");

	// Find the pixels that are candidates for shadow pixels
	buffer->createBuffer( w, h, 1, DEPTH_8U, "shadowcandidates" );
	buffer->get("input_lum")->lessThan(buffer->get("background_lum"), buffer->get("shadowcandidates"));
	buffer->get("shadowcandidates")->and_(buffer->get("smallblobsremoved"));

	// Detect shadow pixels.
	buffer->createBuffer( w, h, 1, DEPTH_8U, "shadow" )->setZero();
	findShadowPixels("shadowcandidates", "shadow");

	// Shadow pixels are now marked in the buffer "shadow".
	// Remove shadows by masking with the inverse shadow buffer.
	buffer->createBuffer( w, h, 1, DEPTH_8U, "foreground" );
	buffer->get("foreground")->copy(buffer->get("smallblobsremoved"));
	buffer->get("shadow")->not_(buffer->get("tmp"));
	buffer->get("foreground")->and_(buffer->get("tmp"));

	buffer->get("foreground")->dilate(structElement, 2);
	buffer->get("foreground")->erode(structElement, 1);
	// Remove any shadowcandidates that were rejected.
	buffer->get("foreground")->not_(buffer->get("tmp"));
	buffer->get("shadow")->and_(buffer->get("tmp"));

	// Label the blobs and add to the blobList.
	buffer->createBuffer( w, h, 1, DEPTH_8U, "labeled" );
	labelBlobs("foreground", "labeled");

	// Pseudo-color the blobs
	buffer->createBuffer( w, h, 3, DEPTH_8U, "pseudo" );
	pseudoColorBlobs("labeled", "pseudo");

	drawDebugOutput();
	PROFILE_END_FUNC;
}
void Bully::ModulePreprocess_HSV::findShadowPixels(const char *src, const char *dst){
	PROFILE_BEGIN_FUNC;
	Buffer* srcbuffer = buffer->get(src);
	Buffer* dstbuffer = buffer->get(dst);
	int w = srcbuffer->w();
	int h = srcbuffer->h();

	if( !buffer->exists("background_hsv") ){
		// This happens only the first iteration.
		buffer->createBuffer( w, h, 3, DEPTH_8U, "background_hsv" );
		buffer->get("background_rgb")->rgb2hsv(buffer->get("background_hsv"));
	}
	Buffer* b_hsv = buffer->get("background_hsv");
	buffer->createBuffer( w, h, 3, DEPTH_8U, "input_hsv" );
	Buffer* i_hsv = buffer->get("input_rgb")->rgb2hsv(buffer->get("input_hsv"));

	float lumlowthreshold = (float)params->get("prepShadowHSVLumLower");
	float lumhighthreshold = (float)params->get("prepShadowHSVLumUpper");
	int huethreshold = (int)params->get("prepShadowHSVHueUpper");
	int satthreshold = (int)params->get("prepShadowHSVSatUpper");

	uchar * srcpixels = srcbuffer->data();
	uchar * ipixels = i_hsv->data();
	uchar * bpixels = b_hsv->data();
	uchar * dstpixels = dstbuffer->data();

	for( int y = 0; y < h; y++ ){
		for( int x = 0; x < w; x++ ){
			if( srcpixels[y*w+x] == 0 ) continue; // Not a candidate shadow pixel

			float vratio = float(ipixels[3*(y*w+x)+2])/float(bpixels[3*(y*w+x)+2]);
			int sdiff = (int)(ipixels[3*(y*w+x)+1]) - (int)(bpixels[3*(y*w+x)+1]);
			int hdiff = abs((int)(ipixels[3*(y*w+x)+0]) - (int)(bpixels[3*(y*w+x)+0]));

			if( (vratio > lumlowthreshold && vratio < lumhighthreshold) &&
				(sdiff < satthreshold) &&
				(hdiff < huethreshold) )
			{
				dstpixels[y*w+x] = 0xFF;
			}
		}
	}
	PROFILE_END_FUNC;
}
void Bully::ModulePreprocess_HSV::drawDebugOutput(){
	ModulePreprocess::drawDebugOutput();
}
