/****************************************************************************

LIMITATIONS:

The recovery from a split is not good as it is.
It works as follows: for every new blob after the group has split, a
difference is computed between the blob and every object that took part in
the previous merge. The object with the smallest difference is taken as the
blob's object (and this object is then updated with the blob's data).

The problem with this is it depends on the order the blobs are traversed.
A better way of doing this is to construct a list for each blob containing
an entry for each object with the difference between the blob and that
object. It could then be sorted on the difference, from best to worst.

Then let each blob have its preferred object as long as no other blob wants
the same object and with a lower difference.

****************************************************************************/
#include "bullymoduletrack_master.h"
#include "bullytypes.h"
#include "bullybloblist.h"
#include "bullyobjectlist.h"
#include "bullyparams.h"
#include "bullymatchmat.h"
#include "bullybuffer.h"
#include "bullybuffermanager.h"
#include <time.h>
#include <vector>
#include <list>
#include <algorithm>
#include <profiling.h>
#include "eriksul.h"
using namespace std;


Bully::ModuleTrack_Master::ModuleTrack_Master( BufferManager &buffer, Params &params )
: ModuleTrack(buffer, params){
	objects = new ObjectList();
	oldObjects = new ObjectList();
}
Bully::ModuleTrack_Master::~ModuleTrack_Master(){
	delete objects; objects = 0;
	delete oldObjects; oldObjects = 0;
}

void Bully::ModuleTrack_Master::process( BlobList *blobList ){
	PROFILE_BEGIN_FUNC;

	int txtOutput = (int)params->get("trackTxtOutput");
	int gfxOutput = (int)params->get("trackGfxOutput");

	// Create the matching matrices (see velastin01).
	int curN = blobList->count();
	int prevN = objects->count();

	MatchMat curM(curN,prevN);	// from current to previous
	MatchMat prevM(prevN,curN);	// from previous to current

	// Match previous to current
	for( int i = objects->first(); i < objects->count(); i++, objects->next() ){
		Object & obj = objects->current();

		for( int j = 0; j < curN; j++ ){
			Blob &blob = *(blobList->get(j));

			if( obj.predictBbox.overlaps( blob.bbox ) )
				prevM(i,j) = 1;
		}
	}

	// Match current to previous
	for( int i = 0; i < curN; i++ ){
		Blob &blob = *(blobList->get(i));

		for( int j = objects->first(); j < objects->count(); j++, objects->next() ){
			Object & obj = objects->current();

			if( obj.predictBbox.overlaps( blob.bbox ) )
				curM(i,j) = 1;
		}
	}

//	prevM.print("prevM");
//	curM.print("curM");

	// Construct the matching strings.
	MatchStr prevS(prevM);
	MatchStr curS(curM);

//	prevS.print("prevS");
//	curS.print("curS");

	// Find the cases: enter, exit, merge, split, one-to-one.
	match( blobList, prevS, curS );

	// Commit the changes using the information from the matching algorithm:

	// ENTER AND MERGE: Add any blobs that have entered the scene.
	for( int i = 0; i < blobList->count(); i++ ){
		Blob &b = *(blobList->get( i ));
		if( b.entered || b.merged ){
			Object newobj(b);
			b.obj = objects->add( newobj );
		}
	}

	// UPDATE
	for( int i = objects->first(); i < objects->count(); i++, objects->next() ){
		Object &obj = objects->current();
		if( obj.matchResult == Object::UPDATE )
			obj.update( *(blobList->get( obj.matchData.matchedToBlob )) );
	}

	// SPLIT AND MERGE
	for( int i = objects->first(); i < objects->count(); i++, objects->next() ){
		Object &obj = objects->current();

		if( obj.matchResult == Object::SPLIT ){

			// Mark this object for deletion.
			obj.matchResult = Object::EXIT;

			// Copy the existing object (the old "group") into a new object,
			// and store it in the oldObjects list for use in the history.
			Object *oldObj = oldObjects->add( obj );

			vector<Object*> objsBeforeMerge = obj.history.involvedObjects;
			vector<Object*> objsAfterSplit;
			vector<int> blobsAfterSplit = obj.matchData.splitIntoBlobs;
			vector<bool> objTaken(objsBeforeMerge.size(), false);

			for( int bi = 0; bi < (int)(blobsAfterSplit.size()); bi++ ){
				// Create an object from the blob
				Blob & b = *(blobList->get( blobsAfterSplit[bi] ));
				Object newObj( b );

				// Put obj in as the only involved object in this split.
				newObj.history.eventType = History::SPLIT;
				newObj.history.involvedObjects.push_back( oldObj );
				oldObj->incRefCount();

				// Figure out which object before the split is the best match.
				if( objsBeforeMerge.size() > 0 ){

					// --------------------------------------------------
					// NB! See comment at top of file.
					// --------------------------------------------------
					int minIndex = getBeforeMergeObj(b, obj, objsBeforeMerge, objTaken);

					objTaken[minIndex] = true;
					newObj.beforeMerge = objsBeforeMerge[minIndex];
				}

				// Finally add the new object to the list of tracked objects.
				objects->add( newObj );
			}

		}else if( obj.matchResult == Object::MERGE ){

			// Mark this object for deletion.
			obj.matchResult = Object::EXIT;

			Object *oldObj = oldObjects->add( obj );

			// Get the new blob/object resulting from the merge.
			Blob &b = *(blobList->get( obj.matchData.mergedIntoBlob ));
			Object *newobj = b.obj;
			assert( b.merged && newobj );

			// Put obj in as an involved object in this merge.
			newobj->history.eventType = History::MERGE;
			newobj->history.involvedObjects.push_back( oldObj );
			oldObj->incRefCount();
		}
	}

	objects->deleteExited();
	oldObjects->deleteUnreferenced();

	drawDebugOutput();

	if( txtOutput > 0 ){
		printf("old(%i) ", oldObjects->count());
		printf("\n");
	}

	PROFILE_END_FUNC;
}
double Bully::ModuleTrack_Master::getBeforeMergeObjDiff(Blob &b, Object &group, Object &o){
// --------------------------------------------------
// NB! See comment at top of file.
// --------------------------------------------------
	double diff = .0;

#if 0 // Histogram
	double histdiff = .0;
	for( int i = 0; i < 256; i++ ){
		histdiff = fabs(b.hist[i] - o.hist[i]);
		diff += histdiff*histdiff;
	}
#endif
#if 1 // Predicted position of object o.
	int mergeFrames = static_cast<int>(group.trajectory.size());
	Vector2i predpos = o.centroid() + o.imgSpeed * mergeFrames;

	diff = (predpos - b.centroid).abs();
/*
	buffer->rect(	predpos.x - o.currentBbox.w/2, predpos.y - o.currentBbox.h/2,
					o.currentBbox.w, o.currentBbox.h,
					0, 0, 0xFF );
*/
#endif
	return diff;
}
int Bully::ModuleTrack_Master::getBeforeMergeObj(Blob &b, Object &group, vector<Object*> &objsBeforeMerge, vector<bool> &objTaken){
// --------------------------------------------------
// NB! See comment at top of file.
// --------------------------------------------------
	double minDiff = 9999999.0;
	int minIndex = 0;

	for( int obji = 0; obji < (int)(objsBeforeMerge.size()); obji++ ){
		if( objTaken[obji] ) continue;

		double diff = getBeforeMergeObjDiff( b, group, *objsBeforeMerge[obji] );
		if( diff < minDiff ){
			minDiff = diff;
			minIndex = obji;
		}
	}
	return minIndex;
}
void Bully::ModuleTrack_Master::match( BlobList *blobList, MatchStr & prevS, MatchStr & curS ){
	PROFILE_BEGIN_FUNC;

	int prevN = prevS.count();
	int curN = curS.count();

	bool ignoreSplit = false;
	bool ignoreMerge = false;
	if( prevN == curN ){
		ignoreSplit = ignoreMerge = true;
	}

	int txtOutput = (int)params->get("trackTxtOutput");
	int gfxOutput = (int)params->get("trackGfxOutput");

	if( txtOutput > 0 )
		printf( "prevN(%i) curN(%i) ", prevN, curN );

	// Used to check for merging.
	vector<int> merged;
	merged.resize(objects->count(), -1);

	for( int i = 0; i < curN; i++ ){

		// ENTER
		if( curS(i).empty() ){ // if curS = 

			int j = 0;
			for( ; j < prevN; j++ ) if( prevS(j).matchedTo( i ) ) break;

			if( j == prevN  ){ // if prevS(j) != i (for all j)
				Blob & b = *(blobList->get(i));
				b.entered = true;

				if( txtOutput > 0 )
					printf( "enter ");
			}

		// MERGE
		}else if( curS(i).size() > 1 && !ignoreMerge ){

			Blob &b = *(blobList->get( i ));
			b.merged = true;

			for( int j = 0; j < curS(i).size(); j++ ) merged[curS(i)[j]] = i;
		}
	}

	for( int i = objects->first(); i < objects->count(); i++, objects->next() ){
		Object & obj = objects->current();
		obj.matchResult = Object::NONE;

		// EXIT
		if( prevS(i).empty() ){ // if prevS(i) = 

			int j = 0;
			for( ; j < curN; j++ ) if( curS(j).matchedTo( i ) ) break;
	
			if( j == curN  ){ // if curS(j) != i (for all j)
				obj.matchResult = Object::EXIT;
				
				if( txtOutput > 0 )
					printf( "exit " );
			}

		}else if( prevS(i).size() == 1 ){

			// ONE-2-ONE
			if( merged[i] < 0 ){
				int j = prevS(i)[0];
				if( (curS(j).size() == 1) && (curS(j)[0] == i) ){
					obj.matchResult = Object::UPDATE;
					obj.matchData.matchedToBlob = j;

					if( txtOutput > 1 )
						printf( "update " );
				}

			// MERGE
			}else{
				obj.matchResult = Object::MERGE;
				obj.matchData.mergedIntoBlob = merged[i];

				if( txtOutput > 0 )
					printf( "merge(%i) ", merged[i] );
			}

		// SPLIT
		}else if( prevS(i).size() > 1 ){

			if( !ignoreSplit ){
				obj.matchResult = Object::SPLIT;
				obj.matchData.splitIntoBlobs = prevS(i);
				if( txtOutput > 0 )
					printf( "split " );
		}else{
				// If this split should be ignored, choose the closest of the blobs.
				double minDist = 9999999.0;
				int minDistIndex = 0;

				for( int spliti = 0; spliti < prevS(i).size(); spliti++ ){
					Blob &b = *(blobList->get(spliti));
					double dist = (b.centroid - obj.centroid()).abs();
					if( dist < minDist ){
						minDist = dist;
						minDistIndex = prevS(i)[spliti];
					}
				}
				obj.matchResult = Object::UPDATE;
				obj.matchData.matchedToBlob = minDistIndex;

				if( txtOutput > 0 )
					printf( "ignoresplit " );
			}

		}else{
			printf( "\n\nERROR: no match was found. This should not happen!\n\n" );
			assert(0);
		}
	}
	PROFILE_END_FUNC;
}
void Bully::ModuleTrack_Master::drawTrajectory_own( Object & obj, uchar r, uchar g, uchar b ){
	if( r==0 && g==0 && b==0 ){
		r = obj.col[0];
		g = obj.col[1];
		b = obj.col[2];
	}

	Buffer *output = buffer->get("output");
	for( int i = 0; i < static_cast<int>(obj.trajectory.size()) - 1; i++ ){
		output->line(	obj.trajectory[i].x, obj.trajectory[i].y,
						obj.trajectory[i+1].x, obj.trajectory[i+1].y,
						r, g, b );
	}
}
void Bully::ModuleTrack_Master::drawTrajectory_all( Object & obj ){
	drawTrajectory_own( obj );

	if( !obj.history.involvedObjects.empty() ){

		vector<Object*>::iterator hobj = obj.history.involvedObjects.begin();
		for( ; hobj != obj.history.involvedObjects.end(); hobj++ ){

			// Only draw trajectories that have not yet been drawn.
			if( !(**hobj).visited ){
				drawTrajectory_all( **hobj );
				(**hobj).visited = true;
			}

			buffer->get("output")->line(
				obj.trajectory[0].x, obj.trajectory[0].y,
				(**hobj).trajectory.back().x, (**hobj).trajectory.back().y,
				obj.col[0], obj.col[1], obj.col[2] );

		}
	}
}
void Bully::ModuleTrack_Master::drawTrajectory_mostlikely( Object & obj, uchar r, uchar g, uchar b ){
	if( r==0 && g==0 && b==0 ){
		r = obj.col[0];
		g = obj.col[1];
		b = obj.col[2];
	}

	drawTrajectory_own( obj, r,g,b );

	if( obj.beforeMerge ){
		Object &bo = *obj.beforeMerge;

		drawTrajectory_mostlikely( bo, r,g,b );

		buffer->get("output")->line(
			obj.trajectory[0].x, obj.trajectory[0].y,
			bo.trajectory.back().x, bo.trajectory.back().y,
			r, g, b );
	}
}
void Bully::ModuleTrack_Master::drawDebugOutput(){
	if( (int)params->get("trackGfxOutput") > 0 ){
		drawPredBoxes();

		if( (int)params->get("trackGfxOutput") > 1 ){
			oldObjects->resetVisited();

			for( int i = objects->first(); i < objects->count(); i++, objects->next() )
				drawTrajectory_all( objects->current() );
		}
	}
}
void Bully::ModuleTrack_Master::drawBoxes(){
	for( int i = objects->first(); i < objects->count(); i++, objects->next() ){
		Object & o = objects->current();
		Recti & r = o.currentBbox();
		buffer->get("output")->rect( r.x, r.y, r.w, r.h, 255, 255, 255 );
	}
}
void Bully::ModuleTrack_Master::drawPredBoxes(){
	for( int i = objects->first(); i < objects->count(); i++, objects->next() ){
		Object & o = objects->current();
		Recti & r = o.predictBbox;
		buffer->get("output")->rect( r.x, r.y, r.w, r.h, 128, 128, 128 );
	}
}
