#include "bullymoduletrack_reading.h"
#include "reading/bullyreadingregiontracker.h"
#include "reading/bullyreadinginputs.h"
#include "reading/bullyreadingresults.h"
#include "reading/bullyreadingtrackedobjectset.h"
#include "reading/bullyreadinggrey8image.h"
#include "reading/bullyreadinglist.h"
#include "reading/tracker_defines_types_and_helpers.h"
#include "bullybuffer.h"
#include "bullybuffermanager.h"
#include "bullyparams.h"
#include "bullybloblist.h"
#include "bullyobjectlist.h"

Bully::ModuleTrack_Reading::ModuleTrack_Reading(BufferManager &buffer, Params &params)
: ModuleTrack(buffer,params) {
	readingRegionTracker = new Reading::RegionTracker;
	inputs = new Reading::Inputs;
	results = new Reading::Results;
	objectList = new ObjectList;
}
Bully::ModuleTrack_Reading::~ModuleTrack_Reading(){
	delete readingRegionTracker;
	delete inputs;
	delete results;
	delete objectList;
}
void Bully::ModuleTrack_Reading::process( BlobList *blobList ){
	++_frameCount;

	// To see how Reading::RegionTracker should be used, see the functions:
	// Tracking.cc: extract_motion_data (puts regions from motion detector in tracked_objects before trackers are called)
	// Tracking.cc: run_trackers (uses among others the RegionTracker)
	// Camera.cc: do_processing (uses Tracking.cc: run_trackers)
	//    notice: results->get_tracked_objects()->remove_old_tracks(...).

	// Set motion_image
	Buffer *bullyMotionBuffer = buffer->get("foreground");
	int motionw = bullyMotionBuffer->w();
	int motionh = bullyMotionBuffer->h();
	unsigned char *motiondata = bullyMotionBuffer->data();
	Reading::frame_id_t frameId = _frameCount;
	Reading::frame_time_t frameTime = 10;

	Reading::Grey8Image readingMotionImage(motionw, motionh, _frameCount, frameTime, motiondata);
	readingMotionImage.set_frame_id(_frameCount);
	results->set_motion_image(&readingMotionImage);

	// Add the blobs to tracked_objects
	for( int blobIndex = 0; blobIndex < blobList->count(); blobIndex++ ){
		Blob *blob = blobList->get(blobIndex);
		int xlo = blob->bbox.x;
		int xhi = blob->bbox.x + blob->bbox.w - 1;
		int ylo = blob->bbox.y;
		int yhi = blob->bbox.y + blob->bbox.h - 1;
		Reading::Image *regimage = readingMotionImage.get_subimage(xlo, xhi, ylo, yhi);
		Reading::Region *reg = new Reading::Region(xlo, xhi, ylo, yhi, regimage);
		reg->source = Reading::MEASUREMENT;
		reg->is_visible = true;
		results->get_tracked_objects()->add_observation( reg, frameId, frameTime );
	}

	results->status = Reading::RESULTS_UNDEFINED;

	// Run region tracker.
	readingRegionTracker->process_frame(inputs, results);
	readingRegionTracker->post_process_frame(inputs, results);

	// check whether all objects have valid (>0) object ids, create if not.
	results->get_tracked_objects()->check_and_create_ids();

	// set the frame id which Results correspond to
	results->frame_id_of_results = frameId;
	
	// now remove old tracks and those that were marked for deletion
	int max_keep_alive = 2; // configuration variable from Camera.cc.
	results->get_tracked_objects()->remove_old_tracks(frameId, max_keep_alive);

	results->status = Reading::RESULTS_COMPLETE;

	updateTrackedObjects();
	removeOldTrackedObjects();
	debugOutput();

	++_frameCount;
}
void Bully::ModuleTrack_Reading::updateTrackedObjects(){
	Reading::ListNode<Reading::TrackedObject> *objnode;
    for( objnode = results->get_tracked_objects()->start(); objnode != NULL; objnode = objnode->next ){
		Reading::TrackedObject *object = objnode->dat;

		if( !object->is_visible ) continue;
		assert( object->regions->no_items == 1 );

	    Reading::ListNode<Reading::Region> *regnode;
		for( regnode = object->regions->first; regnode != NULL; regnode = regnode->next ){
			Reading::Region *region = regnode->dat;

			if( !region->is_visible ) continue;
			if( region->source != Reading::MEASUREMENT ) continue;
			if( region->frame_last_detected != _frameCount ) continue;

			TrackedObject &trackedObj = trackedObjects[object->id];
			trackedObj.id = object->id;
			trackedObj.lastFrameDetected = _frameCount;
			trackedObj.dir.x = region->direction.x;
			trackedObj.dir.y = region->direction.y;
			trackedObj.trajectory.push_back( Vector2i(region->origin.x, region->origin.y) );
		}
	}
}
void Bully::ModuleTrack_Reading::removeOldTrackedObjects(){
	std::map<int, TrackedObject>::iterator objit = trackedObjects.begin();
	while( objit != trackedObjects.end() ){
		std::map<int, TrackedObject>::iterator nextobj = objit;
		nextobj++;

		TrackedObject &obj = objit->second;
		if( obj.lastFrameDetected < _frameCount )
			trackedObjects.erase(objit);

		objit = nextobj;
	}
}
void Bully::ModuleTrack_Reading::handleInput( bool down, char ascii ){
	if( ascii == 'f' ){
		buffer->get("foreground")->show();
	}else if( ascii == 'r' ){
	
	}
}
void Bully::ModuleTrack_Reading::debugOutput(){
	Buffer *output = buffer->get("output");

	std::map<int, TrackedObject>::iterator objit = trackedObjects.begin();
	for( ; objit != trackedObjects.end(); objit++ ){
		TrackedObject &obj = objit->second;

		// Draw trajectory.
		std::vector<Vector2i>::iterator p1 = obj.trajectory.begin();
		std::vector<Vector2i>::iterator p2 = p1;
		p2++;
		for( ; p2 != obj.trajectory.end(); p1++, p2++ ){
			output->line(p1->x, p1->y, p2->x, p2->y, obj.col[0],obj.col[1],obj.col[2]);
		}

		// Draw direction.
		/*
		int cx = obj.trajectory.back().x;
		int cy = obj.trajectory.back().y;
		output->line(cx, cy, cx+obj.dir.x, cy+obj.dir.y, 255, 255, 255);
		*/
	}

	std::cout << "TRACK: regcount(" << trackedObjects.size() << ")" << std::endl;
}
