/*****************************************************************************

	Basic type definitions for the bully system.

	Erik Sundnes Lovlie, 11. November 2005

*****************************************************************************/
#pragma once

#include <math.h>

namespace Bully {

enum PixelDepth {
	DEPTH_8U = 8,
	DEPTH_32F = 32,
};

const double PI = 3.1415926535897932384626433832795;
const float PIf = 3.1415926535897932384626433832795f;

template <typename T>
class Vector2 {
public:
	Vector2(){ x=y=0; }
	Vector2( T x, T y ){ this->x = x; this->y = y; }
	Vector2( const Vector2<T> & v ){ this->x = v.x; this->y = v.y; }
	Vector2 & operator= (const Vector2<T> &v) {
		if( this == &v ) return *this;
		x = v.x;
		y = v.y;
		return *this;
	}
	inline T & operator[]( int i ){ return v[i]; }
	inline T & operator()( int i ){ return v[i]; }

	inline Vector2<T> operator+ (const Vector2<T> &v) const { return Vector2<T>(x+v.x, y+v.y); }
	inline Vector2<T> operator- (const Vector2<T> &v) const { return Vector2<T>(x-v.x, y-v.y); }
	inline Vector2<T> operator* (const T &d) const { return Vector2<T>(x*d, y*d); }
	inline Vector2<T> operator/ (const T &d) const { return Vector2<T>(x/d, y/d); }

	inline bool operator== (const Vector2<T> &v) const { return (v.x == x) && (v.y == y); }
	inline bool operator!= (const Vector2<T> &v) const { return (v.x != x) || (v.y != y); }

	inline double abs() const { return sqrt(static_cast<double>(x*x + y*y)); }
	inline double angle() const { return PI + atan2(y,x); }

	inline Vector2<double> toDouble() const { return Vector2<double>(static_cast<double>(x), static_cast<double>(y)); }
	inline Vector2<int> toInt() const { return Vector2<int>(static_cast<int>(x), static_cast<int>(y)); }

	inline Vector2<T> norm (){
		double l = abs();
		x /= l;
		y /= l;
		return *this;
	}

	union {
		struct{ T x, y; };
		struct{ T v[2]; };
	};
};
typedef Vector2<int> Vector2i;
typedef Vector2<float> Vector2f;
typedef Vector2<double> Vector2d;


template <typename T>
class Vector3 {
public:
	Vector3(){ x=y=z=0; }
	Vector3( T x, T y, T z ){ this->x = x; this->y = y; this->z = z; }
	Vector3( const Vector2<T> & v, T z ){ this->x = v.x; this->y = v.y; this->z = z; }
	Vector3( const Vector3<T> & v ){ this->x = v.x; this->y = v.y; this->z = v.z; }
	Vector3 & operator= (const Vector3<T> &v) {
		if( this == &v ) return *this;
		x = v.x;
		y = v.y;
		z = v.z;
		return *this;
	}
	inline T & operator[]( int i ){ return v[i]; }
	inline T & operator()( int i ){ return v[i]; }

	inline Vector3<T> operator+ (const Vector3<T> &v) const { return Vector3<T>(x+v.x, y+v.y, z+v.z); }
	inline Vector3<T> operator- (const Vector3<T> &v) const { return Vector3<T>(x-v.x, y-v.y, z-v.z); }
	inline Vector3<T> operator* (const T &d) const { return Vector3<T>(x*d, y*d, z*d); }
	inline Vector3<T> operator/ (const T &d) const { return Vector3<T>(x/d, y/d, z/d); }

	inline bool operator== (const Vector3<T> &v) const { return (v.x == x) && (v.y == y) && (v.z == z); }
	inline bool operator!= (const Vector3<T> &v) const { return (v.x != x) || (v.y != y) || (v.z != z); }

	inline double abs() const { return sqrt(static_cast<double>(x*x + y*y + z*z)); }

	inline Vector2<T> to2D() const { return Vector2<T>(x, y); }
	inline Vector3<double> toDouble() const { return Vector3<double>(static_cast<double>(x), static_cast<double>(y), static_cast<double>(z)); }
	inline Vector3<int> toInt() const { return Vector3<int>(static_cast<int>(x), static_cast<int>(y), static_cast<int>(z)); }

	inline Vector3<T> norm (){
		double l = abs();
		x /= l;
		y /= l;
		z /= l;
		return *this;
	}

	union {
		struct{ T x, y, z; };
		struct{ T r, g, b; };
		struct{ T v[3]; };
	};
};
typedef Vector3<int> Vector3i;
typedef Vector3<float> Vector3f;
typedef Vector3<double> Vector3d;
typedef Vector3<unsigned char> Color;


template <typename T>
class Matrix3 {
public:
	Matrix3(){ memset(m, 0, sizeof(T)*3*3); }
	Matrix3( const Matrix3<T> & mat ){ memcpy(m, mat.m, sizeof(T)*3*3); }
	Matrix3 & operator= (const Matrix3<T> &mat) {
		if( this == &mat ) return *this;
		memcpy(m, mat.m, sizeof(T)*3*3);
		return *this;
	}
	inline T & operator()( int r, int c ){ return m[r][c]; }

	inline Matrix3<T> operator* (const T &d) const {
		Matrix3<T> a;
		for( int r = 0; r < 3; r++ )
			for( int c = 0; c < 3; c++ )
				m[r][c] *= d;
		return a;
	}
	inline Vector3<T> operator* (const Vector3<T> &v) const {
		Vector3<T> a;
		for( int r = 0; r < 3; r++ )
			for( int c = 0; c < 3; c++ )
				a.v[r] += m[r][c] * v.v[c];
		return a;
	}
	inline Matrix3<T> operator* (const Matrix3<T> &mat) const {
		Matrix3<T> a;
		for( int r = 0; r < 3; r++ )
			for( int c = 0; c < 3; c++ )
				for( int k = 0; k < 3; k++ )
					a.m[r][c] += m[r][k] * mat.m[k][c];
		return a;
	}

	bool isIdentity() const {
		for( int r = 0; r < 3; r++ ){
			for( int c = 0; c < 3; c++ ){
				if( r == c && m[r][c] != 1 ) return false;
				if( r != c && m[r][c] != 0 ) return false;
			}
		}
		return true;
	}

	T m[3][3];
};
typedef Matrix3<unsigned char> Matrix3b;
typedef Matrix3<int> Matrix3i;
typedef Matrix3<float> Matrix3f;
typedef Matrix3<double> Matrix3d;


template <typename T>
class Rect {
public:
	Rect(){
		x=y=w=h=0;
	}
	Rect( T x, T y, T w, T h ){
		this->x = x;
		this->y = y;
		this->w = w;
		this->h = h;
	}
	Rect( const Rect<T> & r ){
		this->x = r.x;
		this->y = r.y;
		this->w = r.w;
		this->h = r.h;
	}
	Rect & operator= (const Rect<T> &r) {
		if( this == &r ) return *this;
		x = r.x;
		y = r.y;
		w = r.w;
		h = r.h;
		return *this;
	}

	inline bool overlaps( Rect<T> & r ) const {
		return !((x+w <= r.x) || (x >= r.x+r.w) || (y+h <= r.y) || (y >= r.y+r.h));
	}
	inline bool inside( int px, int py ) const {
		return( px >= x && px < x+w && py >= y && py < y+h );
	}
	inline Vector2<T> center() const {
		return Vector2<T>(x+w/2, y+h/2);
	}

	T x, y, w, h;
};
typedef Rect<int> Recti;
typedef Rect<float> Rectf;
typedef Rect<double> Rectd;

} // namespace Bully


#include <ostream>

template<class T>
std::ostream& operator<<(std::ostream& os, Bully::PixelDepth const &pd){
	if( pdf == Bully::DEPTH_8U )
		return os << "DEPTH_8U";
	if( pdf == Bully::DEPTH_32F )
		return os << "DEPTH_32F";
	return os << pd;
}

template<class T>
std::ostream& operator<<(std::ostream& os, Bully::Vector2<T> const &v){
	return os << "(" << v.x << "," << v.y << ")";
}

template<class T>
std::ostream& operator<<(std::ostream& os, Bully::Vector3<T> const &v){
	return os << "(" << v.x << "," << v.y << "," << v.z << ")";
}

template<class T>
std::ostream& operator<<(std::ostream& os, Bully::Rect<T> const &r){
	return os << "(" << r.x << "," << r.y << "," << r.w << "," << r.h << ")";
}

template<class T>
std::ostream& operator<<(std::ostream& os, Bully::Matrix3<T> const &m){
	for( int r = 0; r < 3; r++ ){
		for( int c = 0; c < 3; c++ ){
			os << m.m[r][c] << "\t";
		}
		os << endl;
	}
	return os;
}
